#!/usr/bin/env python3
import os
import json
import sys
from urllib.request import urlopen
from urllib.parse import urlparse, parse_qs, urlencode, quote
from http.server import BaseHTTPRequestHandler, HTTPServer

# find instances at searx.space
#searx = 'searx.xyz'
searx = 'searx.prvcy.eu'
#searx = 'search.mdosch.de'
#searx = 'nibblehole.com'

policy = {} # maps label to 1 or -1
labels = {} # maps url to label to source to value (nested dicts)
labels1 = {} # maps source to (url, label) to value
sources = { 'me' : 0 } # maps source to tier
tmax = 0 # maximum tier of configured sources

def log(*args, **kwargs):
	print(*args, file=sys.stderr, **kwargs)

def reputation(j):
	if sources[j] == 0:
		return 1
	n = 0
	d = 0
	for (i, k), v in labels1[j].items():
		e = expectation(i, k, sources[j] - 1)
		if e > 0.5:
			x = 1
		elif e < 0.5:
			x = 0
		else:
			continue
		n += abs(v - x)
		d += 1
	if d == 0:
		return 0
	return max(1 - 2*n/d, 0)

def expectation(i, k, t):
	if t < 0 or not i in labels or not k in labels[i]:
		return 0.5
	e = expectation(i, k, t - 1)
	if e != 0.5:
		return e
	n = 0
	d = 0
	for j, v in labels[i][k].items():
		if sources[j] != t:
			continue
		n += reputation(j) * labels[i][k][j]
		d += reputation(j)
	if d == 0:
		return 0.5
	return n/d

def adjustment(result):
	r = 1
	for i, v in policy.items():
		e = expectation(result['url'], i, tmax)
		if v == 1:	# attr i is favored
			f = e
		else:		# attr i is disfavored
			f = 1 - e
		if f >= 0.5:
			r *= 2*f
		else:
			r *= 0.5 / (1 - f)
		# padding e for pretty formatting
		e = round(e, 2)
		estr = str(e)
		estr += '&nbsp;' * (6 - len(estr))
		result['labels'] += f'E({i})&nbsp;=&nbsp;{estr} '
	return r

def readpolicy(fname, v):
	try:
		with open(fname, 'r') as f:
			for line in f:
				l = line.strip()
				if l:
					policy[l] = v
	except FileNotFoundError as e:
		log(e)

def setpolicy(s, v):
	for line in s.splitlines():
		l = line.strip()
		if l:
			policy[l] = v

def sync(s, fname):
	if s is not None:
		with open(fname, 'w') as f:
			if len(s) and s[-1] != '\n':
				s += '\n'
			f.write(s)
	else:
		try:
			with open(fname, 'r') as f:
				s = f.read()
		except FileNotFoundError as e:
			log(e)
			s = ''
	return s

def do_search(qs):
	q = ''
	if qs:
		qdict = parse_qs(qs)
		if 'q' in qdict:
			q = qdict['q'][0]
	results = []
	jdict = None
	if q:
		if q == 'privacy':
			try:
				with open('privacy.json', 'r') as f:
					jdict = json.load(f)
			except FileNotFoundError as e:
				log(e)
		if not jdict:
			url = f'https://{searx}/search?format=json&{urlencode({"q": q}, doseq=True)}'
			response = urlopen(url)
			jdict = json.loads(response.read())
		results = jdict['results']
	for r in results:
		r['labels'] = ''
		r['ascore'] = r['score'] * adjustment(r)
		r['score'] = round(r['score'], 3)
		r['ascore'] = round(r['ascore'], 3)
	results = sorted(results, key = lambda r: r['ascore'], reverse = True)

	policyhtml = ''
	for l in policy:
		if policy[l] > 0:
			sign = '+'
		else:
			sign = '&minus;'
		policyhtml += f'<tr><td>{l}</td><td>{sign}</td></tr>\n'

	sourcehtml = ''
	for j in sources:
		if j != 'me':
			sourcehtml += f'<tr><td>{j}({sources[j]})</td><td>{round(reputation(j), 4)}</td></tr>\n'

	resulthtml = ''
	for r in results:
		qurl = quote(r['url'], safe='')
		resulthtml += f'<li style="margin: 0 0 10px 0"><a href="view?u={qurl}"><b>{r["title"]}</b></a><table style="font-family: monospace;">\n'
		for k in ['url', 'labels', 'score', 'ascore']:
			resulthtml += f'<tr><td valign="top">{k}:&nbsp;</td><td>{r[k]}</td></tr>\n'
		resulthtml += '</table></li>\n'

	return f'''<html><head><meta charset="utf-8"/><body>
<div style="float: right; padding: 8px; background: white;">
<table border=1 width="100%"><tr><th colspan=2>Policy <a href="policy">(edit)</a></th></tr>
{policyhtml}
</table><br>
<table border=1 width="100%"><tr><th colspan=2>Label sources</th></tr>
{sourcehtml}
</table></div>
<h2>PURESearch</h2>
<form action="search" method="GET">
<input type="TEXT" name="q" value="{q}">
<input type="SUBMIT" value="Search">
</form>
<hr>
<ol>
{resulthtml}
</ol>
</body></html>
'''

def do_policy(qs):
	saved = False
	promote = demote = None
	if qs:
		qdict = parse_qs(qs, keep_blank_values=True)
		if 'p' in qdict:
			promote = qdict['p'][0]
			saved = True
		if 'd' in qdict:
			demote = qdict['d'][0]
			saved = True
	promote = sync(promote, 'config/promote')
	demote = sync(demote, 'config/demote')
	policy = {}
	setpolicy(promote, 1)
	setpolicy(demote, -1)

	return f'''<html><body><form action="policy" method="GET">
Labels to <b>promote</b>, one per line:<br>
<textarea rows="10" cols="40" name="p">{promote}</textarea>
<br><br>
Labels to <b>demote</b>, one per line:<br>
<textarea rows="10" cols="40" name="d">{demote}</textarea>
<br><br>
<input type="submit" value="Save" />
</form>
<a href="search">Back to search</a>
</body>
</html>
'''

def do_view(qs):
	u = ''
	if qs:
		qdict = parse_qs(qs)
		if 'u' in qdict:
			u = qdict['u'][0]

	return f'''<html>
<frameset cols = "*, 0">
<frame src="{u}">
<frame src="label?{qs}">
</frameset>
</html>
'''

def do_label(qs):
	url = ''
	label = value = None
	if qs:
		qdict = parse_qs(qs)
		if 'u' in qdict:
			url = qdict['u'][0]
		if 'l' in qdict:
			label = qdict['l'][0]
		if 'v' in qdict:
			value = qdict['v'][0]

	if url and label and value:
		if url not in labels:
			labels[url] = {}
		if label not in labels[url]:
			labels[url][label] = {}
		labels[url][label]['me'] = value
		labels1['me'][(url, label)] = value
		try:
			with open('labels/me', 'a') as f:
				f.write(f'{label}\t{value}\t{url}\n')
		except FileNotFoundError as e:
			log(e)

	labelhtml = ''
	if url in labels:
		for l in labels[url]:
			labelhtml += f'<li><b>{l}</b>:<ul>'
			for j, v in labels[url][l].items():
				labelhtml += f'<li>{j}({sources[j]}): {v}</li>'
			labelhtml += '</ul></li>'
	else:
		labelhtml = '<li>n/a</li>'

	return f'''<html>
<form action="label" method="GET">
<table width="100%">
<tr><td>URL:</td><td width="100%"><input style="width: 100%" type="TEXT" name="u" value="{url}" required></td></tr>
<tr><td>Label:</td><td width="100%"><input style="width: 100%" type="TEXT" name="l" required></td></tr>
<tr><td>Value:</td><td width="100%"><label><input type="radio" name="v" value="1" required>True</label><br>
<label><input type="radio" name="v" value="-1" required>False</label></td></tr>
</table>
<input type="submit" value="Submit">
</form>
<hr>Current labels for {url}:<br><ul>{labelhtml}</ul></html>
'''

class handler(BaseHTTPRequestHandler):
	def do_GET(self):
		o = urlparse(self.path)
		m = ''
		if o.path == '/' or o.path == '/search':
			m = do_search(o.query)
		elif o.path == '/policy':
			m = do_policy(o.query)
		elif o.path == '/view':
			m = do_view(o.query)
		elif o.path == '/label':
			m = do_label(o.query)
		if m:
			self.send_response(200)
			self.send_header('Content-type','text/html')
			self.end_headers()
		else:
			self.send_response(404)
			self.send_header('Content-type','text/html')
			m = "Not Found"
		self.wfile.write(bytes(m, 'utf8'))

readpolicy('config/promote', 1)
readpolicy('config/demote', -1)
try:
	with open('config/sources') as f:
		for line in f:
			fields = line.strip().split('\t')
			t = int(fields[1])
			if t > tmax:
				tmax = t
			sources[fields[0]] = t
except FileNotFoundError as e:
	log(e)

for j in sources:
	try:
		with open('labels/' + j) as f:
			for line in f:
				fields = line.strip().split('\t')
				l, v, u = fields[0], fields[1], fields[2]
				if not u in labels:
					labels[u] = {}
				if not l in labels[u]:
					labels[u][l] = {}
				labels[u][l][j] = int(v)
				if not j in labels1:
					labels1[j] = {}
				labels1[j][(u,l)] = int(v)
	except FileNotFoundError as e:
		log(e)

with HTTPServer(('', 8000), handler) as server:
	server.serve_forever()
